# $Id: jefflib26.py,v 1.35 2012-10-03 01:19:47 jdeifik Exp $
# Copyright Jeff turbo Deifik 2008
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This is for python 2.6..2.999
# Dec-11-2008	JTD	Added class addend
# Jan-06-2009	JTD	Added UKKA_Dist_With_Max
# Jul-13-2009	JTD	Added Progress_Indicator
# Nov-24-2011	JTD	5 times more resolution for Bar_Count


from __future__ import print_function
import os			# os.rename
import stat			# stat.S_ISDIR
import sys			# sys.exit
import hashlib			# md5sum
from array import array


# SORTING FUNCTIONS

# The bubblesort algorithm.  Returns a sorted list.
# Don't use this on large lists, roughly 50+ elements
def Bubble_Sort(lis, cmp = None):
    if cmp == None: cmp = Default_Cmp
    end = len(lis) -1

    while end > 0:		# While we have more to do.
        tmp = 0
        for i in range(0, end):
            res = cmp(lis[i], lis[i+1])
            if res > 0:			# Switch elements.
                (lis[i], lis[i+1]) = (lis[i+1], lis[i])
                tmp = i
        end = tmp		# Array is totally sorted up to here.

    return lis


# The shakersort algorithm.  Faster than bubblesort. Returns a sorted list.
# Don't use this on large lists, roughly 50+ elements
def Shaker_Sort(lis, cmp = None):
    if cmp == None: cmp = Default_Cmp
    begin = 0
    end = len(lis) -1

    while True:				# While we have more to do.
        tmp = begin
        for i in range(0, end):
            res = cmp(lis[i], lis[i+1])
            if res > 0:			# Switch elements.
                (lis[i], lis[i+1]) = (lis[i+1], lis[i])
                tmp = i
        if tmp == begin: break		# while loop
        end = tmp			# Array is totally sorted up to here.

        for i in range(end-1, begin-1, -1):
            res = cmp(lis[i], lis[i+1])
            if res > 0:			# Switch elements.
                (lis[i], lis[i+1]) = (lis[i+1], lis[i])
                tmp = i
        if tmp == end: break		# while loop
        begin = tmp			# Array is totally sorted up to here.

    return lis


# The insertion sort algorithm. Order n^2. Returns a sorted list.
def Insertion_Sort(lis, cmp = None):
    if cmp == None: cmp = Default_Cmp

    nel = len(lis)			# Number of elements

    for i in range(1, nel):
        item = lis[i]
        iHole = i
# keep moving the hole to next smaller index until lis[iHole - 1] is <= item
        while iHole > 0 and cmp(lis[iHole - 1],item) > 0:
            # move hole to next smaller index
            lis[iHole] = lis[iHole - 1]
            iHole -= 1
        # put item in the hole
        lis[iHole] = item
    return lis


# The shellsort algorithm. Order n^4/3. Returns a sorted list.
def Shell_Sort(lis, cmp = None):
    # The values in gaps are from Marcin Ciura, see wikipedia on shellsort
    gaps = [1070178, 428071, 171228, 68491, 27396, 10958, 4383,
            1753, 701, 301, 132, 57, 23, 10, 4, 1]

    if cmp == None: cmp = Default_Cmp

    nel = len(lis)			# Number of elements

    for gap in gaps:
        for i in range(gap, nel):
            temp = lis[i]
            j = i
            while j >= gap and cmp(lis[j - gap], temp) > 0:
                lis[j] = lis[j-gap]	# base[j] <- base[j-gap]
                j -= gap
            lis[j] = temp
    return lis


# The mergesort algorithm.  Returns a sorted list. Calls merge.
def Merge_Sort(lis, cmp = None):
    if cmp == None: cmp = Default_Cmp
    a_len = len(lis)
    if a_len <= 1:
        return lis
    else:
        m = (a_len + 1) // 2
        b_list = Merge_Sort(lis[0:m], cmp)
        c_list = Merge_Sort(lis[m:], cmp)
        return _Merge(b_list, c_list, cmp)


# Merge two sorted lists, used by Merge_Sort
def _Merge(lis_a, lis_b, cmp = None):
    ret = []
    a, b = 0, 0

    len_a = len(lis_a)
    len_b = len(lis_b)
    while a < len_a and b < len_b:
        res = cmp(lis_a[a], lis_b[b])
        if res <= 0:
            ret.append(lis_a[a])
            a += 1
# The 'else' can't be res >= 0, because we need a stable sort.
# We must exhaust all of a, before adding any b if the cmp==0.
        else:
            ret.append(lis_b[b])
            b += 1
    if a < len_a:
        ret.extend(lis_a[a:])
    if b < len_b:
        ret.extend(lis_b[b:])
    return ret


# The bottom up mergesort algorithm.  Returns a sorted list.
# Unlike the common merge sort, this is not recursive.
def Bottom_Merge_Sort(A, cmp = None):
    if cmp == None: cmp = Default_Cmp
    n = len(A)
    if n <= 1:
        return A

    # B is a work array needed for Bottom_Merge_Sort
    B = list(A)		# B is a copy of A, but all different memory

    # each 1-element run in A is already "sorted".
 
    # Make successively longer sorted runs of length 2, 4, 8, 16...
    # until whole array is sorted
    # for (width = 1; width < n; width = 2 * width)
    width = 1
    while width < n:
        # array A is full of runs of length width
        for i in range(0, n, 2 * width):
            # merge two runs: A[i:i+width-1] and A[i+width:i+2*width-1] to B[]
            # or copy A[i:n-1] to B[] ( if(i+width >= n) )
            _BottomUpMerge(A, i, min(i+width, n), min(i+2*width, n), B, cmp)
 
        # now work array B is full of runs of length 2*width
        # copy array B to array A for next iteration
        #   a more efficient implementation would swap the roles of A and B
        for i in range(0, n):
            A[i] = B[i]
        # now array A is full of runs of length 2*width
        width *= 2

    return A


def _BottomUpMerge(A, iLeft, iRight, iEnd, B, cmp):
    i0 = iLeft
    i1 = iRight
 
    # while there are elements in the left or right lists
    for j in range(iLeft, iEnd):
        # if left list head exists and is <= existing right list head
        if i0 < iRight and (i1 >= iEnd or cmp(A[i0],A[i1]) <= 0):
            B[j] = A[i0]
            i0 += 1
        else:
            B[j] = A[i1]
            i1 += 1


# The quicksort algorithm.  Returns a sorted list. Calls quick. NOT STABLE!
def Quick_Sort(lis, cmp = None):
    if cmp == None: cmp = Default_Cmp
    Quick(lis, 0, len(lis) - 1, cmp)
#    print('quick_sort, lis is ', lis)
    return lis


# Used by Quick_Sort, does the real work of quicksort
def Quick(lis, low, high, cmp):

    if high <= low: return		# End of recursion

    if high - low == 1:			# Special case, 2 elements
        if cmp(lis[low], lis[high]) > 0:
            (lis[low], lis[high]) = (lis[high], lis[low])

#    print('debug, top of quick is low ', low, 'high', high, 'lis', lis[low:high+1])
# choose pivot
    i = (high+low)//2
    if cmp(lis[low], lis[i]) > 0: (lis[low], lis[i]) = (lis[i], lis[low])
    if cmp(lis[i], lis[high]) > 0: (lis[i], lis[high]) = (lis[high], lis[i])
    if cmp(lis[low], lis[i]) > 0: (lis[low], lis[i]) = (lis[i], lis[low])
# Now lis[low] <= lis[i] <= lis[high], and lis[i] is now the pivot.
    assert(lis[low] <= lis[i])
    assert(lis[i] <= lis[high])

    (lis[low], lis[i]) = (lis[i], lis[low])	# Move pivot to lis[high]
    pivot = lis[high]
#    print('debug, pivot ', pivot)

    sto = low					# Partition
    for i in range(sto, high):
        if cmp(lis[i], pivot) < 0:
             (lis[i], lis[sto]) = (lis[sto], lis[i])
             sto += 1
    (lis[high], lis[sto]) = (lis[sto], lis[high])	# Move pivot
#    print('debug, moved pivot ', lis[low:high+1])
#    for i in range(low, sto): assert(lis[i] < lis[sto])
#    for i in range(sto, high+1): assert(lis[sto] <= lis[i])

# recursive sorts
#    print('debug going to call (', low, sto-1, '), (', sto+1, high, ')');
    Quick(lis, low, sto-1, cmp)
    Quick(lis, sto+1, high, cmp)


# used by various sorts
def Default_Cmp(a, b):
    if a > b: return 1
    elif a < b: return -1
    else: return 0


# LIST FUNCTIONS

# Uniqueifles a list, based on adjacent elements
def Uniq(lis, cmp = None):
    if cmp == None: cmp = Default_Cmp
    new_lis = []

    len_lis = len(lis)

    if len_lis > 0:
        new_lis = [ lis[0] ]		# Initialization
        new_elem = lis[0]

        for i in range(1, len_lis):
            if cmp(new_elem, lis[i]) == 0:
                continue
            else:
                new_lis.append(lis[i])
                new_elem = lis[i]

    return new_lis


# Print usage message and exit
def Usage(liszt):
    for s in liszt:
        print(s, file=sys.stderr)
    sys.exit(1)


# Counts non-zero elements in a list
def Pop_Count(liszt):
    pop = 0
    for l in liszt:
        if l: pop += 1
    return pop


# Counts val elements in a list
def Pop_Count_Val(liszt, val):
    pop = 0
    for l in liszt:
        if l == val: pop += 1
    return pop


# Return min value in a list
def Min_Value(liszt):
    minn = liszt[0]
    for e in liszt:
        if e < minn:
            minn = e
    return minn


# Return max value in a list
def Max_Value(liszt):
    maxx = liszt[0]
    for e in liszt:
        if e > maxx:
            maxx = e
    return maxx


# Return sum of all elements in a list
def Sum_List(liszt):
    sum = 0
    for e in liszt:
        sum += e
    return sum


# Given an array or list, initialize the first siz elements to 0
def Init_List(lis, siz):
    for i in range (0, siz):
        lis.append(0)


# DATE FUNCTIONS

# Convert string month to numeric month, Jan->1.. Dec->12
def Month_Str_To_Num(mon_str):
   dic = {"jan": 1, "feb": 2, "mar": 3, "apr": 4, "may": 5, "jun": 6,
          "jul": 7, "aug": 8, "sep": 9, "oct": 10, "nov": 11, "dec":12}

   l_mon = mon_str.lower()
   if l_mon in dic: return dic[l_mon]
   else:
       print('Unknown month', mon_str, file=sys.stderr)
       raise ValueError


# Convert numeric month to string month, 1->Jan.. 12->Dec
def Month_Num_To_Str(mon_num):
   mon_list = ("", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")

   if (mon_num >=1 and mon_num <= 12):
       return mon_list[mon_num]
   else:
       print(' Unknown month', mon_num, file=sys.stderr)
       raise ValueError


# Validate date
def Validate_Date(year, mon, day):
    if int(mon) < 1 or int(mon) > 12:
        print('Month out of range', mon, day, year, file=sys.stderr)
        raise ValueError
    if int(day) < 1 or int(day) > 31:
        print('Day out of range', mon, day, year, file=sys.stderr)
        raise ValueError
    if int(year) < 1 or int(year) > 10000:
        print('Year out of range', mon, day, year, file=sys.stderr)
        raise ValueError


# FILE FUNCTIONS

def Try_To_Link(the_link, the_file, verbose):
# Try to link
    if verbose: print("Linking", the_link, "to", the_file)
    try:
        os.link(the_link, the_file)
    except Exception as ex:
        print("Failed to link", the_link, "to", the_file)
        print('exception = ', ex)
        return False
    return True


def Try_To_Rename(name, new_name, verbose):
# Try to rename
    if verbose: print("Renaming", name, "to", new_name)
    try:
        os.rename(name, new_name)
    except Exception as ex:
        print("Failed to rename", name, "to", new_name)
        print('exception = ', ex)
        return False
    return True


def Try_To_Remove(name, verbose):
# Try to remove
    if verbose: print("Removing", name)
    try:
        os.remove(name)
    except Exception as ex:
        print("Failed to remove", name)
        print('exception = ', ex)
        return False
    return True


# Try to make a directory
def Try_Mkdir(the_dir):
    try:
        os.mkdir(the_dir)
    except Exception as ex:
        print('Failed to mkdir', the_dir)
        print('exception = ', ex)


# Try to call open, if fail print message and exit
def Vopen(*args):
# Try to open
    try:
        fp = open(*args)
        return(fp)
    except Exception as ex:
        print('Failed to open', args)
        print('exception = ', ex)
        sys.exit(1)


# Take in list of directories, Create a table of files indexed by size.
# Files smaller than min_size will be ignored.
# Progress is logged with the progress function if progress_size > 0
# If ignore_links is True, ignore hard links
# Return table, which is of type dictionary.
# Each table element is a tuple (file_name, inode)
def files_and_sizes(dir_list, min_size = 0, progress = None,
                    ignore_links = False):
    table = {}

    if progress != None:
        print('\nReading directories ', dir_list)
    while len(dir_list) > 0:
        try:
            names = os.listdir(dir_list[0])
        except:
            dir_list.pop(0)
            continue

        for name in names:
            if dir_list[0] == '/':
                fname = '/' + name
            else:
                fname = dir_list[0] + '/' + name
            try:
                s = os.stat(fname)
            except:
                continue

            if (progress != None):
                progress()

            if stat.S_ISDIR(s[stat.ST_MODE]):
                dir_list.append(fname)
            elif stat.S_ISREG(s[stat.ST_MODE]):
                num_links = s[stat.ST_NLINK]
                link_ok = ignore_links == False or num_links == 1
                size = s[stat.ST_SIZE]
                inode = s[stat.ST_INO]
                if link_ok and size > min_size:
                    if size in table:
                        table[size].append((fname, inode))
                    else:
                        table[size] = [(fname, inode)]

        dir_list.pop(0)
    return table


# Return a list of directories (full path)
# as well as a list of plain files (no path) in directory.
def Plain_Files_In_Directory(cur_dir):
    plain_files = []
    dir_list = []

    for n in os.listdir(cur_dir):
        full_n = cur_dir + '/' + n
        try:
            s = os.stat(full_n)
        except:
            continue

        if stat.S_ISDIR(s[stat.ST_MODE]):
            dir_list.append(full_n)
        elif stat.S_ISREG(s[stat.ST_MODE]):
            plain_files.append(n)
    return(dir_list, plain_files)


# Take in a file, open it, compute a md5sum,
# return a string of md5sum and file name
# Cribbed from python/tools/scripts/md5sum.py
def Md5sum_On_File(file_name, base_name):
    BLOCKSIZE = 1024*1024

    try:
        fp = open(file_name, "rb")
        m = hashlib.md5()
        while 1:
            block = fp.read(BLOCKSIZE)
            if not block:
                break
            m.update(block)
        fp.close()
        return (m.hexdigest() + " *" + base_name)
    except Exception as ex:
        return ("failed compute_md5sum_on_file, " + base_name + "exception = " +
                ex.__str__())


# Take in a file, open it, compute a md5sum, on first BLOCKSIZE bytes
# return a 32 bit integer of the md5sum leading 32 bits
def Md5sum_On_First_Block_Of_File(file_name):
    BLOCKSIZE = 512

    try:
        fp = open(file_name, "rb")
        m = hashlib.md5()
        block = fp.read(BLOCKSIZE)
        if block:
            m.update(block)
        fp.close()
        return int(m.hexdigest()[0:8], 16)
    except Exception as ex:
        return ""


# Take in a file, open it, read it into a list,
# return the list
def File_To_List(file_name):
    res = []
    fp = open(file_name)
    res = fp.readlines()
    fp.close()
    return res


# Take in a file name and a list, copy the list to the file,
# ignoring whatever used to be in the file.
def List_To_File(file_name, the_list):
    try:
        fp = open(file_name, "wb")

        for lstr in the_list :
            fp.write(lstr)
        fp.close()
    except:
        return ("failed list_to_file, can't open " + file_name)


# FILE COUNTING FUNCTIONS

# Process one file, counting each char in the file
# Input is an opened file
# Output is a tuple of and an array of ints of size 256,
# the highest non-zero value index, and the highest value
def File_Count_Chars(f):
    c_data = array('L')		# Where the char counts are stored

# There is likely a more elegant way to create an array of uints, but it
# eludes me for now...
    Init_List(c_data, 256)

    while 1:
        line = f.readline()
        if line == '' or line == b'': break
        for c in line:
            c_data[ord(c)] += 1

    # Find index of highest character
    c_high = 255
    while (c_data[c_high] == 0):
        c_high -= 1

    # Find the largest count
    c_max = 0
    for i in range(0,256):
        if c_data[i] > c_max:
            c_max = c_data[i]

    return (c_data, c_high, c_max)


# Print a general bar graph, via ascii art
# data is an array of integers
# high is the highest index in the array whose valuse is non-zero
def Bar_Count(data, high):
    height = 20			# Height of graph
    width = 64			# Width of graph
                                # need to change bottim line code if it changes...
    height_plus = height + 1
    count = array('L')		# Where the bar graph data is stored

# There is likely a more elegant way to create an array of uints, but it
# eludes me for now...
    Init_List(count, width)

    quanta = 1 + (high//width)

    for a in range(0,width):	# Get array of groups of quanta values.
        b = a * quanta
        for b in range (a*quanta, (a+1)*quanta):
            if b <= high:
                count[a] += data[b]

    maxx = Max_Value(count)
    if maxx % height == 0:		# Find the graph scaling.
        scale = maxx//height
    else:
        scale = (maxx//height) + 1

    print()				# Print a blank line.

    for a in range(1, height_plus):	# Display the bargraph.
        print('{0:8d}  '.format(((height_plus)-a)*scale), end='')
        for b in range (0,width):
            if count[b] >= (height_plus-a)*scale:
                sys.stdout.write('X')
                continue
# Five times the resolution of just using X and x
            Found = False
            for z in range(9, 0, -1):
                zf = float(z)/10.0
                if count[b] >= (height + zf - a)*scale:
                    sys.stdout.write(str(z))
                    Found = True
                    break
            if Found == False:
                sys.stdout.write(' ')
        print()

    print("          |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |   |")
    if high >= 1000:
        sys.stdout.write('          0       ')
        for i in range (1,9):
            sys.stdout.write('{0:<8d}'.format((quanta*8*i)-1))
        print()
    else:
        sys.stdout.write('          0   ')
        for i in range (1,17):
            sys.stdout.write('{0:<4d}'.format((quanta*4*i)-1))
        print()


# Print an array of integers, in a tabular fashion
# data is an array of integers
# high is the highest index in the array whose valuse is non-zero
def Display_Count(data, high):
    len_d = len(data)

    maxx = Max_Value(data)
    x = len(str(maxx)) + 2
    fmt = str(x) + 'd'			# Derive format string
    print()				# Print a blank line.
    for i in range(0, high+1, 8):	# Print the array

        non_zero = False
        for j in range(i, min(i+8, len_d)):
            if data[j] != 0:
                non_zero = True
                break

        if non_zero:
            print('{0:03d}..{1:03d}'.format(i, i+7), end='')

            for j in range(i, min(i+8, len_d)):
                print(' {0:{1}}'.format(data[j], fmt), end='')
            print()


# BYTES FUNCTIONS

def String_to_Bytes(stir):
    liszt = []				# List of integers
    for c in stir:
        liszt.append(ord(c))
    return bytes(liszt)


def Has_High_Byte(bites):
    for b in bites:
        if ord(b) > 127: return(True)
    return(False)


# Dunno what will happen if called with high byte on
#def Bytes_to_String(bites):
#    liszt = []				# List of characters
#    for b in bites:
#        liszt.append(chr(b))
#    return(List_to_String(liszt))


# MISC FUNCTIONS

# Takes in opened file to read, and calls func on each line read.
# Will work with either binary or text opening files
def Process_Line(fp, func):
    while 1:
        data = fp.readline()
        if data == '' or data == b'': return
        func(data)


# Print an error message, and exit with failed status
def Print_Error(*stir):
    print(*stir, file=sys.stderr)
    sys.exit(1)


# SIMPLE CLASSES

# A Progress Indicator

class Progress_Indicator:

    # Create the progrss indicator
    # period is how often to make output
    # mode 0 makes dots
    # mode 1 makes twirling lines
    def __init__(self, period, mode = 0):
        self.__period = period
        self.__count = 0
        self.__mode = mode
        self.__twirl = 0	# For the twirling lines

    # Reset progress info
    def Reset(self):
        self.__count = 0
        self.__twirl = 0

    # Write progress info
    def P_I(self):
        self.__count += 1
        if (self.__count % self.__period == 0):
            if self.__mode == 0:
                sys.stdout.write('.') ; sys.stdout.flush()
            if self.__mode == 1:
                if   self.__twirl == 0: st = '\\'
                elif self.__twirl == 1: st = '\b|'
                elif self.__twirl == 2: st = '\b/'
                elif self.__twirl == 3: st = '\b-'
                self.__twirl += 1
                self.__twirl = self.__twirl % 4
                sys.stdout.write(st) ; sys.stdout.flush()

# Take in strings to write, and insert a newline every num outputs.
# Separate the strings by sep, write output to out

class Wrap:

    # Create the wrap class
    def __init__(self, num, sep=' ', out=sys.stdout):
        self.__num = num
        self.__sep = sep
        self.__out = out
        self.__count = 0

    # Write a newline, unless it was done at last output.
    def Flush(self):
        if self.__count != 0:
            print(file=self.__out)
            self.__count = 0

    # Write output
    def Output(self, stir):
        if self.__count == 0:
            print(stir, end='', file=self.__out)
        else:
            print(self.__sep + stir, end='', file=self.__out)
        self.__count += 1
        if self.__count == self.__num:
            print(file=self.__out)
            self.__count = 0


# Class to append strings to the end of a file.
# Before any writing to a file is done, a call to create the object via
# __init__ must be made, setting various parameters. Then calls to APPEND_STR
# can be made. APPEND_FLUSH will make sure that a close is done, to ensure
# data gets out to disk. If these routines are used, then other means of
# opening, and writing  to the file should not be used at the same time.
# There are several reasons for the existance of these functions.
# One reason is to minimize the opening and seeking to the end of files
# for performance reasons.
# Written by Jeff Deifik Dec-30-1993 (c), Dec-11-2008 (python)

class Addend:

    # Create the append class
    def __init__(self, count, fname):
        self.__count_writes = count
        self.__name = fname[:]
        self.__fptr = None
        self.__internal_count = 0

    def Add(self, stir):
        # Open file if it happens to be closed.
        if self.__internal_count == 0:
            self.__fptr = Vopen(self.__name, 'a')
        print(stir, file=self.__fptr)	# Do the write.
        self.__internal_count += 1	# Increment the internal counter.
        # If its time to flush, do the flush.
        if self.__internal_count == self.__count_writes:
            self.Flush()

    def Flush(self):
        if self.__internal_count > 0:
            self.__fptr.close()
            self.__internal_count = 0
